/*
 * Parser according to syntax specification in PROTOCOL_INPUT_SPEC
 * treats the input file as if containing only one specification for a protocol
 * and gives a sst/9 structure as output
 */
% read_protocol_file(+FileName, -InitTree) :-	InitTree is sst/9, read from file are the principal list and initial intruder knowledge needed for the init state,
%												rest is filled in with default values.
read_protocol_file(FileName, InitTree, PropertyPairs) :-
	parse_file(FileName, PL, IIKTokens, PPL),
	parse_term_list(IIKTokens, IIK),
	parse_ppl(PPL, PropertyPairs),
	%writef("iik: %w\n", [IIK]),
	create_principal_node_list(PL, PNL),
	InitTree = sst(0, 0, state(PNL, IIK, []), root, root, [], nn, nn, []),
	writeln('\nCreated initial node successfully.').
read_properties(FileName, PropertyPairs) :-
	parse_file(FileName, _, _, PPL),
	parse_ppl(PPL, PropertyPairs),
	writeln('\nCreated only property pairs.').

% create_principal_node_list(+PrincipalTokenList, -PrincipalList) :- construct each principal used in the protocol.
create_principal_node_list([], []).
create_principal_node_list([PT|PTs], PNL) :-
	create_principal_node(PT, 1, PN),
	create_principal_node_list(PTs, RestPN),
	append([PN], RestPN, PNL).

% create_principal_node(+TokenList, +Num, -PrincipalNode) :- construct a p/1 for each Element in +TokenList that matches [+Num|_]
create_principal_node(TL, Num, PN) :-
	findall(X, (member(X, TL), X=[Num|_]), XL),
	create_principal_successors(TL, XL, PSL),
	PN = p(PSL).

% create_principal_successors(+TokenList, +PrincipalTupleList, -PrincipalSuccessorList)
create_principal_successors(_, [], []).
create_principal_successors(TL, [X|Ts], PSL) :-
	X=[_,_,N,_|Rest1],
	create_principal_node(TL, N, P),
	parse_next_term(Rest1, LHS, [_,_|Rest2]),
	parse_next_term(Rest2, RHS, _),
	PS = ps(rule(LHS,RHS), P),
	create_principal_successors(TL, Ts, RestPSL),
	append([PS], RestPSL, PSL).

% parse_file(+FileName, -PrincipalList, -IntruderKnowledge)
parse_file(FileName, DefPList, SIK, PropPairs) :-
	read_complete_file(FileName, TokenList),
	parse_token_list(TokenList, PList, IK, PropPairs, Defines),
	substitute_defines(Defines, Defines, SubDefines),
	substitute_terms(IK, SubDefines, SIK),
	substitute_principals(PList, SubDefines, DefPList),
	(\+ var(PList); !, writeln('Error: no principals specified'), fail),
	(\+ var(IK); !, writeln('Error: no intruder knowledge specified'), fail), !.

read_complete_file(FileName, TokenList) :-
	open(FileName, read, InStream),
	set_input(InStream),
	read_file(TokenList),
	close(InStream).

% read_file(-TokenList) :- Inputstream is already set up appropriately
read_file([]) :-
	at_end_of_stream, !.
read_file(TokenList) :-
	readln(Token),
	read_file(Rest),
	append([Token], Rest, TokenList).

% parse_token_list(+TokenList, -PrincipalList, -InitialIntruderKnowledge)
parse_token_list([], _, _, _, _) :- !.
% empty line
parse_token_list([[]|TLs], PL, IIK, PP, D) :-
	parse_token_list(TLs, PL, IIK, PP, D).
% begin reading new principal
parse_token_list([Principal|TLs], FinRTL, IIK, PP, D) :-
	principal_token(Principal),
	read_rule_token_list(TLs, Rem, RTL),
	parse_token_list(Rem, TmpRTL, IIK, PP, D),
	((\+ var(TmpRTL),!,append([RTL], TmpRTL, FinRTL)); FinRTL=[RTL]).
% begin reading intruder knowledge
parse_token_list([IK|TLs], PL, IIK, PP, D) :-
	ik_token(IK),
	read_intruder_knowledge(TLs, Rem, IIK),
	parse_token_list(Rem, PL, _, PP, D), !.
% begin reading property pairs
parse_token_list([Prop|TLs], PL, IIK, PP, D) :-
	prop_token(Prop),
	read_prop_pairs(TLs, Rem, PP),
	parse_token_list(Rem, IIK, PL, _, D).
parse_token_list([Def|TLs], PL, IIK, PP, Defines) :-
	define_token(Def),
	read_defines(TLs, Rem, Defines),
	parse_token_list(Rem, PL, IIK, PP, _).

% read_intruder_knowledge(+TokenList, -TokenListRemainder, -IIK) :- return all lines (token-wise) until eof or beginning of another principal
read_intruder_knowledge([], [], []) :- !.
read_intruder_knowledge([IK|TLs], [IK|TLs], []) :-
	ik_token(IK), !.
read_intruder_knowledge([PT|TLs], [PT|TLs], []) :-
	principal_token(PT), !.
read_intruder_knowledge([PPT|TLs], [PPT|TLs], []) :-
	prop_token(PPT), !.
read_intruder_knowledge([DEF|TLs], [DEF|TLs], []) :-
	define_token(DEF), !.
read_intruder_knowledge([T|TLs], Rem, IIK) :-
	read_intruder_knowledge(TLs, Rem, IK2),
	append(T, IK2, IIK).

% read_rule_token_list(+TokenList, -TokenListRemainder, -RuleTokenList)
read_rule_token_list([], [], []) :- !.
read_rule_token_list([IK|TLs], [IK|TLs], []) :-
	ik_token(IK), !.
read_rule_token_list([PT|TLs], [PT|TLs], []) :-
	principal_token(PT), !.
read_rule_token_list([PPT|TLs], [PPT|TLs], []) :-
	prop_token(PPT), !.
read_rule_token_list([DEF|TLs], [DEF|TLs], []) :-
	define_token(DEF), !.
read_rule_token_list([RT|TLs], Rem, RTL) :-
	read_rule_token_list(TLs, Rem, RTL2),
	((RT\==[],!,append([RT], RTL2, RTL)); RTL=RT).

% read_prop_pairs(+TokenList, -TokenRemainderList, -PropPairList)
read_prop_pairs([], [], []) :- !.
read_prop_pairs([IK|TLs], [IK|TLs], []) :-
	ik_token(IK), !.
read_prop_pairs([PT|TLs], [PT|TLs], []) :-
	principal_token(PT), !.
read_prop_pairs([PPT|TLs], [PPT|TLs], []) :-
	prop_token(PPT), !.
read_prop_pairs([DEF|TLs], [DEF|TLs], []) :-
	define_token(DEF), !.
read_prop_pairs([PPT|TLs], Rem, PPL) :-
	read_prop_pairs(TLs, Rem, PPL2),
	((PPT\==[],!,append([PPT], PPL2, PPL)); PPL=PPT).

read_defines([], [], []) :- !.
read_defines([IK|TLs], [IK|TLs], []) :-
	ik_token(IK), !.
read_defines([PT|TLs], [PT|TLs], []) :-
	principal_token(PT), !.
read_defines([PPT|TLs], [PPT|TLs], []) :-
	prop_token(PPT), !.
read_defines([DEF|TLs], [DEF|TLs], []) :-
	define_token(DEF), !.
read_defines([DT|TLs], Rem, DefT) :-
	read_defines(TLs, Rem, DefT2),
	((DT\==[],!,append([DT],DefT2,DefT)); DefT=DefT2).

% check
ik_token(T) :-
	concat_all(T, CA),
	downcase_atom(CA, DcA),
	(DcA=='ik'; DcA=='ik:'; DcA=='intruderknowledge'; DcA=='intruderknowledge:').
principal_token(T) :-
	concat_all(T, CA),
	downcase_atom(CA, DcA),
	(DcA=='principal'; DcA=='principal:').
prop_token(T) :-
	concat_all(T, CA),
	downcase_atom(CA, DcA),
	(DcA=='properties'; DcA=='properties:').
define_token(T) :-
	concat_all(T, CA),
	downcase_atom(CA, DcA),
	(DcA=='defines'; DcA=='defines:').

% use atom_concat/3 to  concatenate a list of unspecified length of atoms
concat_all([A], A).
concat_all([A, B], CA) :- atom_concat(A, B, CA).
concat_all([A, B|R], CA) :-
	R \== [],
	atom_concat(A, B, Tmp),
	concat_all([Tmp|R], CA).

del_bad_atoms([], []).
del_bad_atoms([','|R], RR) :-
	del_bad_atoms(R, RR).
del_bad_atoms([X|R], RR) :-
	del_bad_atoms(R, Tmp),
	append([X], Tmp, RR).

substitute_defines([], _, []) :- !.
substitute_defines([D|Ds], Def, Result) :-
	D=[ID,Sep|Terms],
	sub_def_terms(Terms, Def, DTerms),
	substitute_defines(Ds, Def, DDs),
	append([ID,Sep],DTerms,Tmp),
	(has_def_term(Tmp, Def), !, substitute_defines([Tmp], Def, [ND]); ND=Tmp),
	(DDs\==[], !, Result=[ND|DDs]; Result=[ND]).

sub_def_terms([], _, []) :- !.
sub_def_terms([L|Ts], Def, Result) :-
	is_list(L),
	sub_def_terms(L, Def, LRes),
	sub_def_terms(Ts, Def, TsRes),
	append([LRes], TsRes, Result), !.
sub_def_terms([T|Ts], Def, Result) :-
	member([T,_|Rest], Def),
	sub_def_terms(Ts, Def, TsRest),
	append(Rest, TsRest, Result), !.
sub_def_terms([T|Ts], Def, [T|DTs]) :-
	sub_def_terms(Ts, Def, DTs).

has_def_term([_|Terms], Def) :-
	member(X, Terms),
	member([X|_], Def), !.

substitute_principals([], _, []) :- !.
substitute_principals([P|Ps], Def, [DP|DPs]) :-
	sub_def_terms(P, Def, DP),
	substitute_principals(Ps, Def, DPs).

% substitute_terms(+TermList, +DefinesList, -DefTermList)
substitute_terms([], _, []) :- !.
substitute_terms([D|Ls], Def, [RTerm|SubLs]) :-
	member([D,_,RTerm], Def),
	substitute_terms(Ls, Def, SubLs), !.
substitute_terms([[N,'-',M,':',LHS,'-','>',RHS]|Ls], Def, [RPr|SubLs]) :-
	member([LHS,_|RLHS], Def),
	member([RHS,_|RRHS], Def),
	append([N,'-',M,':'], RLHS, Tmp),
	append(Tmp, ['-','>'|RRHS], RPr),
	substitute_terms(Ls, Def, SubLs), !.
substitute_terms([[N,'-',M,':',LHS,'-','>',RHS]|Ls], Def, [RPr|SubLs]) :-
	member([LHS,_|RLHS], Def),
	append([N,'-',M,':'], RLHS, Tmp),
	append(Tmp, ['-','>',RHS], RPr),
	substitute_terms(Ls, Def, SubLs), !.
substitute_terms([[N,'-',M,':',LHS,'-','>',RHS]|Ls], Def, [RPr|SubLs]) :-
	member([RHS,_|RRHS], Def),
	%append([N,'-',M,':',LHS,'-','>'],RRHS, RPr),
	RPr=[N,'-',M,':',LHS,'-','>'|RRHS],
	substitute_terms(Ls, Def, SubLs), !.
substitute_terms([T|Ls], Def, [T|Res]) :-
	substitute_terms(Ls, Def, Res).


% parse_ppl
parse_ppl(P, P) :- var(P), !.
parse_ppl([P], [PP]) :- parse_pair(P,PP).
parse_ppl([P|TLs], [PP|PPL]) :-
	parse_pair(P, PP),
	parse_ppl(TLs, PPL).
parse_pair(P, PP) :-
	del_bad_atoms(P, DP),
	find_closing_bracket(DP, '(', ')', 0, RestP),
	append(Pos, [')'|RestP], DP), !,
	Pos = [_|RPos],
	RestP = [_|RRestP], append(Neg, [')'], RRestP),
	PP = (RPos, Neg).

% parse_next_term(+CharList, -Term, RestCharList)
parse_next_term([], [], []) :- !.
parse_next_term([','|CLs], T, R) :- parse_next_term(CLs, T, R), !.
parse_next_term(['var',':',T|RestCL], var(T), RestCL) :- !.
parse_next_term(Enc, EncTerm, RestCL) :-
	Enc=['{'|X],
	find_closing_bracket(X, '{', '}', 1, App),
	append(Term, ['}'|App], X),!,
	App = [Type,_|KeyTermRest],
	parse_next_term(Term, ParsedTerm, _),
	parse_next_term(KeyTermRest, ParsedKeyTerm, RestCL),
	(
		(Type==s, EncTerm=sym_enc(ParsedTerm, ParsedKeyTerm));
		(Type==a, EncTerm=asym_enc(ParsedTerm, ParsedKeyTerm));
		(writeln('error: wrong encryption-type identifier'))
	), !.
parse_next_term(Hash, hash(ParsedTerm), RestCL) :-
	Hash=['hash','('|X],
	find_closing_bracket(X, '(', ')', 1, RestCL),
	append(Term, [')'|RestCL], X),!,
	parse_next_term(Term, ParsedTerm, _).
	%writef("X %w\nTerm %w\n
parse_next_term(Sig, sig(ParsedTerm, pk(ParsedKey)), RestCL) :-
	Sig=['sig','('|X],
	find_closing_bracket(X, '(', ')', 1, App),
	append(Term, [')'|App], X),!,
	parse_next_term(Term, ParsedTerm, _),
	App=[':'|KeyTermRest],
	parse_next_term(KeyTermRest, ParsedKey, RestCL).
parse_next_term(SC, sc(S, R, ParsedTerm), RestCL) :-
	SC=['sc','('|X],
	find_closing_bracket(X, '(', ')', 1, RestCL),
	append([S,',',R,','|Term], [')'|RestCL], X), !,
	parse_next_term(Term, ParsedTerm, _).
parse_next_term(Pair, PairTerms, RestCL) :-
	Pair=['<'|X],
	find_closing_bracket(X, '<', '>', 1, RestCL),
	append(TokenList, ['>'|RestCL], X), !,
	parse_term_list(TokenList, PairTerms).
parse_next_term([A|CLs], A, CLs) :- atomic(A), !.

% parse_term_list(+TokenList, -TermList)
parse_term_list([], []).
parse_term_list(TL, [T|RT]) :-
	parse_next_term(TL, T, RTL),
	parse_term_list(RTL, RT), !.

% fcb(+CharList, +BracketOpening, + BracketClosing, +Num, -Rest) :- CharList is assumed to be not empty
find_closing_bracket([],_,_,_,_) :- !.
find_closing_bracket([OBr|CLs], OBr, CBr, N, Rest) :-
	N1 is N + 1,
	find_closing_bracket(CLs, OBr, CBr, N1, Rest).
find_closing_bracket([CBr|Rest], _, CBr, N, Rest) :- N == 1.
find_closing_bracket([CBr|CLs], OBr, CBr, N, Rest) :-
	N1 is N - 1,
	find_closing_bracket(CLs, OBr, CBr, N1, Rest).
find_closing_bracket([_|CLs], OBr, CBr, N, Rest) :-
	find_closing_bracket(CLs, OBr, CBr, N, Rest).

% test
pnt(A,B,C) :- parse_next_term(A,B,C).
ptl(A,B) :- parse_term_list(A,B).
ppf(X,Y) :- principal_token_to_parser_format(X,Y).
ipf(X,Y) :- ik_token_to_parser_format(X,Y).
rcf :- read_complete_file('good', [_|[F]]), retractall(y(_)), assert(y(F)).
rp :- read_protocol_file('good', P), retractall(p(_)), assert(p(P)).
d :-
	set_prolog_flag(toplevel_print_options, [quoted(true), portray(true), max_depth(40), attributes(portray)]),
	set_prolog_flag(debugger_print_options, [quoted(true), portray(true), max_depth(40), attributes(portray)]).
dd(I) :-
	set_prolog_flag(toplevel_print_options, [quoted(true), portray(true), max_depth(I), attributes(portray)]),
	set_prolog_flag(debugger_print_options, [quoted(true), portray(true), max_depth(I), attributes(portray)]).
at(X,Y) :- all_tokens(X,Y).
rb([],[]). %remove bracket
rb([B|R], X) :-
	B=='<',
	append(X, ['>'], R).
fcb(A,B,C,D,E) :- find_closing_bracket(A,B,C,D,E).
fcbc(A, B) :- find_closing_bracket(A, '{', '}', 1, B).
fcbr(A, B) :- find_closing_bracket(A, '(', ')', 1, B).
fcbs(A, B) :- find_closing_bracket(A, '<', '>', 1, B).
nt(A,B,C) :- next_token(A,B,C).
%get defines
gd(File) :-
	read_complete_file(File, TL),
	parse_token_list(TL, PL, _, _, Def),
	retractall(a(_,_)),
	assert(a(PL,Def)), !.
%testme
% <a,sig({<a,b,<j,l>>}:ki):k,c,hash(<y,x>),sig(hash(x)):keyyy>
% <<a,sc(1,2,3),c>,<{x}a:k,b>>
%toy examples:
rp_tb :- read_protocol_file('toyex_bad', P, _), retractall(toybad(_)), assert(toybad(P)).
rp_tg :- read_protocol_file('toyex_good', P, _), retractall(toygood(_)), assert(toygood(P)).
rp_a  :- read_protocol_file('ASW_Alice', P, _), retractall(alicex(_)), assert(alicex(P)).
%principal list from sst
pl(sst(_,_,state(PL,_,_),_,_,_,_,_,_), PL) :- !.

toy_good(VT) :-
	rp_tg,
	toygood(P),
	full_tree(P,T),
	get_variable_tree(T,_,VT).
toy_bad(VT) :-
	rp_tb,
	toybad(P),
	full_tree(P,T),
	get_variable_tree(T,_, VT).

tree(File, VT, PP) :-
	read_protocol_file(File, Pr, PP),
	full_tree(Pr, T),
	get_variable_tree(T, _, VT).



































